"""Module for recording and saving measuremnts.
"""

import re


class ProfileRecord(object):
    """Simple record to track how many times a function/path is called.
    """

    def __init__(self, name, hits):
        """Initializer.

        :param name:   String name of function or stack path.

        :param hits:   Number of times it is called.
        """
        self.name = name
        self.hits = hits


class CountingRecorder(object):
    """Recorder which just counts how many times something is called.
    """

    def __init__(self):
        self.my_db = {}

    def record(self, measurement):
        """Record a measurement.

        :param measurement:     An ox_profile.core.metrics.Meaasurement
                                for profiling the program.

        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        PURPOSE:  This method is called to record a measurement. Different
                  recorders may track different things about a measurement.

        """
        record = self.my_db.get(measurement.name, 0)
        self.my_db[measurement.name] = record + 1

    def query(self, re_filter='.*', max_records=10):
        """Query the database of measurements.

        :param re_filter='.*':      String regular expression for records
                                    to include in query.

        :param max_records=10:      Maximum number of records to include.

        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        :return:   The pair (result, count) where count num_records is the
                   total number of records in the database and result is
                   a list of ProfileRecord instances sorted to start from the
                   record with the most hits to the least with at most
                   max_records included.

        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        PURPOSE:   Query the database.

        """
        regexp = re.compile(re_filter)
        calls = {}
        num_records = len(self.my_db)
        for name, item in self.my_db.items():
            name_list = name.split(';')
            for fname in name_list:
                if regexp.search(fname):
                    calls[fname] = calls.get(fname, 0) + item
        my_hits = list(reversed(sorted(calls.items(),
                                       key=lambda pair: pair[1])))
        result = [ProfileRecord(name, hits) for name, hits in my_hits[
            :max_records]]
        return result, num_records



"""Module for handling measurement and sampling of data.
"""


class Measurement(object):
    """Measurement of profiling information.

    The idea is that you initialize an instance of a Measurement when
    you want to get some profiling information about a program.
    """

    def __init__(self, frame):
        """Initializer.

        :param frame:    Stack frame to measure.

        """
        self.name = self.snap(frame)

    def get_path(self):
        """Return backtrace path for snapped measurement.
        """
        return self.name.split(';')

    def snap(self, frame):
        """Snap a measurement for the given stack frame (called by __init__).

        :param frame:     Stack frame to take a measurement about.

        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        :return:   String representing nicely formatted backtrace of functions
                   for given stack frame.

        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        PURPOSE:   Meant to be called internally by __init__ to snap a
                   measurement.

        """
        stack = []
        self.name = '{}({})'.format(frame.f_code.co_name,
                                    frame.f_globals.get('__name__'))
        while frame is not None:
            formatted_frame = '{}({})'.format(frame.f_code.co_name,
                                              frame.f_globals.get('__name__'))
            stack.append(formatted_frame)
            frame = frame.f_back

        formatted_stack = ';'.join(reversed(stack))
        return formatted_stack



"""Tools to sample running programs.
"""

import logging
import sys

from ox_profile.core import metrics


class Sampler(object):
    """Basic class to sample program for statistical profiling.
    """

    def __init__(self, my_db):
        self.my_db = my_db

    def get_measure_tool(self):
        """Get a class or function to call to take a measurement.

        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        :return:  class or function to call to take a measurement.

        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        PURPOSE:  This method provides the "measurement tool" we are going
                  to use in profiling. Sub-classes could override this to
                  take different kinds of measurements.

        """
        dummy = self
        return metrics.Measurement

    def run(self):
        """Run the sampler to make a measurement of the current stack frames.
        """
        frames_dict = sys._current_frames()  # pylint: disable=protected-access
        measure_tool = self.get_measure_tool()
        for frame_id, frame in frames_dict.items():
            logging.debug('Recording frame with id %s', str(frame_id))
            measure = measure_tool(frame)
            self.my_db.record(measure)

    def __call__(self, *args, **kwargs):
        "Syntactic sugar to call `self.run(*args, **kwargs)`."

        return self.run(*args, **kwargs)



"""Basic views for ox_profile flask UI.
"""

import logging
from functools import wraps

from flask import request, Markup, current_app, render_template, url_for
from flask.ext.login import login_required
from flask.ext.login import current_user

from ox_profile.ui.flask import OX_PROF_BP


def access_problem_p():
    """If user is allowed access to ox_profile, return None else return error.

    This is meant to be called when a user tries to access ox_profile routes
    so that non-privledged users cannot cause trouble. It looks in
    current_app.config['OX_PROF_USERS'] for a dictionary or set of allowed
    users.
    """
    allowed = current_app.config.get('OX_PROF_USERS', {})
    user = current_user.name
    if user in allowed:
        return None

    msg = 'Current user %s not in the %i allowed ox profile users.%s' % (
        user, len(allowed),
        '\nConfigure app.config["OX_PROF_USERS"] if necessary.')
    return render_template('ox_prof_err.html', error_msg=msg)


def restrict_access(my_func):
    """Simple decorator to call access_problem_p before execution.
    """
    @wraps(my_func)
    def wrapper(*args, **kwds):
        """Check access_problem_p and then execute wrapped function.
        """

        problem = access_problem_p()
        if problem is None:
            return my_func(*args, **kwds)
        return problem

    return wrapper


@OX_PROF_BP.route('/')
@OX_PROF_BP.route('/index')
@login_required
@restrict_access
def index():
    """Top-level index for ox profile.
    """
    commands = [(n, url_for('%s.%s' % ('ox_profile', n))) for n in [
        'status', 'pause', 'unpause']]

    return render_template('ox_prof_intro.html', commands=commands)


@OX_PROF_BP.route('/status')
@login_required
@restrict_access
def status():
    """Show status of current profiling.
    """
    re_filter = request.args.get('re_filter', '.*')
    max_records = int(request.args.get('max_records', 50))

    query, total_records = OX_PROF_BP.launcher.sampler.my_db.query(
        re_filter=re_filter, max_records=max_records)

    return render_template(
        'ox_prof_status.html', launcher=OX_PROF_BP.launcher,
        max_records=max_records, total_records=total_records, query=query)


@OX_PROF_BP.route('/pause')
@login_required
@restrict_access
def pause():
    """Pause the profiler.
    """
    OX_PROF_BP.launcher.pause()
    return render_template('ox_prof_msg.html', message='paused')


@OX_PROF_BP.route('/unpause')
@login_required
@restrict_access
def unpause():
    """Unpause profiler or start it for the first time.
    """
    launcher = OX_PROF_BP.launcher
    msgs = []
    if not launcher.is_alive():
        msgs.append('Started thread for first time')
        launcher.start()
    launcher.unpause()
    msgs.append('unpaused')
    return render_template('ox_prof_msg.html', message=Markup(
        '\n<BR>\n'.join(msgs)))


@OX_PROF_BP.route('/set_interval')
@login_required
@restrict_access
def set_interval():
    "Set the sampling interval"
    interval = request.args.get('interval', None)
    problem = None
    try:
        fint = float(interval)
        if fint < 0:
            raise ValueError('Negative interval of %s illogical.')
        if fint > 10:
            raise ValueError('Interval > 10 is useless.')
    except ValueError as problem:
        logging.debug('Could not convert interval %s to float because %s',
                      interval, problem)
    if problem:
        return render_template('ox_prof_err.html', error_msg=problem)
    OX_PROF_BP.launcher.set_interval(fint)

    return render_template('ox_prof_msg.html', message=(
        'Changed sampling interval to %f' % fint))



"""Flask Blueprint for ox_Profile.

This module provides the OX_PROF_BP variable representing the main
ox_profile blueprint. If you are incorporating ox_profile into your own
flask application, you should call register_blueprint(OX_PROF_BP)
appropriately.
"""

import logging

from flask import Blueprint

from ox_profile.core import launchers


class OxProfBlueprint(Blueprint):
    """Subclass flask Blueprint to provide custom blueprint for ox_profile.

    """

    def __init__(self, *args, **kwargs):
        Blueprint.__init__(self, *args, **kwargs)
        self.launcher = launchers.SimpleLauncher()

    def register(self, app, *args, **kwargs):
        """Override default register method to also activate plugins.

        :arg app, *args, **kwargs:  As for usual Blueprint.register method.

        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        :returns:  As usual for Blueprint.register.

        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        PURPOSE:   Override registration so we can start plugins.

        """
        result = Blueprint.register(self, app, *args, **kwargs)
        logging.debug('Registered ox_profile blueprint')
        return result

    @staticmethod
    def get_help():
        "Get some help (point user to github repo)."
        print('Please see github repo for ox_profile more information.')


OX_PROF_BP = OxProfBlueprint(
    'ox_profile', __name__, template_folder='templates',
    static_folder='static', url_prefix='/ox_profile')



"""Some simple and quick tests to run.
"""

import doctest


def simple_doctest():
    """Simple test of whether things work.

>>> import time
>>> from threading import Timer
>>> from ox_profile.core import launchers
>>> launcher = launchers.SimpleLauncher()
>>> launcher.start()
>>> launcher.unpause()
>>> def boring(sleep=4):
...     if sleep <= 1:
...          print('getting unbored')
...          time.sleep(sleep)
...          return False
...     else:
...          print('bored for %.2f more seconds' % sleep)
...          time.sleep(1)
...          return boring(sleep - 1.0)
...
>>> boring()
bored for 4.00 more seconds
bored for 3.00 more seconds
bored for 2.00 more seconds
getting unbored
False
>>> launcher.is_alive()
True
>>> launcher.cancel()
>>> query, total_records = launcher.sampler.my_db.query()
>>> assert len(query) > 3
>>> print('Items in query:%c  - %s' % (10, ('%c  - '%10).join([
...     '%s: %s' % (i.name, i.hits) for i in query]))) # doctest: +ELLIPSIS
Items in query:
  - ...
>>> time.sleep(3)  # Give launcher time to die
>>> launcher.is_alive()
False
"""


if __name__ == '__main__':
    doctest.testmod()
    print('Finished Tests')



"""Module containing profiling launchers.
"""

import time
import logging
import threading


from ox_profile.core import sampling, recording


class SamplingTracker:

    def __init__(self):
        self.calls = 0
        self.wait = 0.0
        self.wait_sq = 0.0

    def reset(self):
        self.calls = 0
        self.wait = 0.0
        self.wait_sq = 0.0

    def snap(self, prev):
        my_wait = time.time() - prev
        self.calls += 1
        self.wait += my_wait
        self.wait_sq += my_wait**2

    def stats(self):
        if self.calls == 0:
            return 'No samples taken'

        mean = self.wait/self.calls
        return {'mean': mean,
                'stdev': (self.wait_sq/self.calls - mean**2)**0.5}

class SimpleLauncher(threading.Thread):
    """Simple profiling launcher.
    """

    def __init__(self, sampler=None, stop_flag=None, interval=.001,
                 *args, **kwargs):
        """Initializer.

        :param sampler=None:   Instance of a profiling sampler such as
                               `ox_profile.core.sampling.Sampler` (which is
                               used as default if sampler is None).

        :param stop_flag=None: Optional threading.Event to use in forcing the
                               thread to stop. If None, then one will be
                               created so you can stop via `self.cancel()`.

        :param interval=.001:  How often (in seconds) to sample the program.

        :param *args, **kwargs:  Passed to threading.Thread.__init__.

        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        PURPOSE:      The profiler is basically running as a daemon thread
                      and continusly

        """
        self.tracker = SamplingTracker()
        self.sampler = sampler if sampler else sampling.Sampler(
            recording.CountingRecorder())
        self.interval = interval
        self.stop_flag = stop_flag if stop_flag else threading.Event()
        self.stop_flag.clear()
        self.unpaused = threading.Event()
        self.pause()
        threading.Thread.__init__(self, *args, **kwargs)
        self.daemon = True
        assert self.isDaemon()

    def set_interval(self, new_interval):
        assert new_interval > 0 and new_interval < 10
        self.interval = new_interval
        self.tracker.reset()

    def pause(self):
        """Pause profiling until self.unpause() is called.
        """
        self.unpaused.clear()

    def unpause(self):
        """Un-pause profiling if paused.

        Note that you must call `self.start()` to initially start and trying
        to unpause before calling `self.start()` will not doo much.  Note this
        is a *thread* so you should call `self.start()` *NOT* `self.run()`.
        """
        self.unpaused.set()

    def is_paused(self):
        """Return whether we are paused or not (see pause/unpause methods).
        """
        return not self.unpaused.is_set()

    def run(self):
        """Start running the profiler.

        *IMPORTANT*:  This is a *thread* so you should call `self.start()`
                      *NOT* `self.run()`. Do *NOT* call self.run directly.
        """
        logging.info('Starting Launcher')
        prev = time.time()
        while not self.stop_flag.is_set():
            time.sleep(self.interval)
            self.tracker.snap(prev)
            self.unpaused.wait()
            prev = time.time()
            self.sampler()

        logging.info('Stopping Launcher')

    def cancel(self):
        """Cancel the running profiler.

        This is useful since the profiler is a thread. Although pause will
        pause the thread, you need to call cancel (and make sure it is
        unpaused) for the thread to actually exit.
        """
        self.stop_flag.set()
